@@ -1,4 +1,5 @@
-##############################################################################
+
+# ##############################################################################
 #
 # Copyright (c) 2006-2008 Zope Corporation and Contributors.
 # All Rights Reserved.
@@ -22,7 +23,7 @@ class Recipe:
     def __init__(self, buildout, name, options):
         self.egg = zc.recipe.egg.Egg(buildout, options['recipe'], options)
         self.buildout, self.options, self.name = buildout, options, name
-        self.zope2_egg = options.get('zope2-egg', False)
+        self.zope2_location = options.get('zope2-location', '')
 
         options['location'] = os.path.join(
             buildout['buildout']['parts-directory'],
@@ -46,13 +47,25 @@ class Recipe:
         # patch the result. A better approach might be to provide independent
         # instance-creation logic, but this raises lots of issues that
         # need to be stored out first.
-        mkzopeinstance = os.path.join(
-            options['zope2-location'], 'bin', 'mkzopeinstance.py')
-        if not os.path.exists(mkzopeinstance):
+        if not self.zope2_location:
             mkzopeinstance = os.path.join(
-                options['zope2-location'], 'utilities', 'mkzopeinstance.py')
-        if sys.platform[:3].lower() == "win":
-            mkzopeinstance = '"%s"' % mkzopeinstance
+                options['bin-directory'], 'mkzopeinstance')
+            if not mkzopeinstance:
+                # EEE
+                return
+
+        else:
+            mkzopeinstance = os.path.join(
+                self.zope2_location, 'bin', 'mkzopeinstance.py')
+            if not os.path.exists(mkzopeinstance):
+                mkzopeinstance = os.path.join(
+                    self.zope2_location, 'utilities', 'mkzopeinstance.py')
+            if sys.platform[:3].lower() == "win":
+                mkzopeinstance = '"%s"' % mkzopeinstance
+        
+        if not mkzopeinstance:
+            # EEE
+            return
 
         assert os.spawnl(
             os.P_WAIT, os.path.normpath(options['executable']),
@@ -94,7 +107,7 @@ class Recipe:
         ws_locations = [d.location for d in ws]
 
         if os.path.exists(location):
-            # See is we can stop. We need to see if the working set path
+            # See if we can stop. We need to see if the working set path
             # has changed.
             saved_path = os.path.join(location, 'etc', '.eggs')
             if os.path.isfile(saved_path):
@@ -102,17 +115,16 @@ class Recipe:
                     '\n'.join(ws_locations)
                     ):
                     # Something has changed. Blow away the instance.
-                    self.install()
+                    return self.install()
                 elif options.get('site-zcml'):
                     self.build_package_includes()
 
             # Nothing has changed.
+            self.install_scripts()
             return location
 
         else:
-            self.install()
-
-        return location
+            return self.install()
 
     def build_zope_conf(self):
         """Create a zope.conf file
@@ -178,6 +190,9 @@ class Recipe:
                                         'off')
                 webdav_address = webdav_server_template % (webdav_address,
                                                            webdav_conn_close)
+            icp_address = options.get('icp-address', '')
+            if icp_address:
+                icp_address = icp_server_template % icp_address
             effective_user = options.get('effective-user', '')
             if effective_user:
                 effective_user = 'effective-user %s' % effective_user
@@ -274,10 +289,26 @@ class Recipe:
                                 "host='%(host)s' password='%(password)s'")
                     rel_storage = dict(dsn=template % rel_storage)
 
+                rel_storage_outer_opts = (
+                    'poll-interval',
+                    'pack-gc',
+                    'pack-dry-run',
+                    'pack-batch-timeout',
+                    'pack-duty-cycle',
+                    'pack-max-delay',
+                    'cache-servers',
+                    'cache-module-name',
+                    )
+
                 opts = dict(
                     type=type_,
-                    opts='\n'.join([' ' * 12 + ' '.join(item)
-                                   for item in rel_storage.iteritems()]))
+                    db_opts='\n'.join(' ' * 12 + ' '.join((k, v))
+                                      for k, v in rel_storage.iteritems()
+                                      if k not in rel_storage_outer_opts),
+                    rs_opts='\n'.join(' ' * 8 + ' '.join((k, v))
+                                      for k, v in rel_storage.iteritems()
+                                      if k in rel_storage_outer_opts),
+                    )
                 storage_snippet = rel_storage_template % opts
 
             else:
@@ -291,7 +322,9 @@ class Recipe:
                 storage_snippet = file_storage_template % file_storage
 
                 blob_storage = options.get('blob-storage', None)
-                demo_storage = options.get('demo-storage', None)
+                demo_storage = options.get('demo-storage', 'off') \
+                             not in ('off', 'disable', 'false')
+
                 if blob_storage and demo_storage:
                     raise ValueError("Both blob and demo storage cannot be used"
                                      " at the same time.")
@@ -412,6 +445,7 @@ class Recipe:
                                         http_fast_listen = http_fast_listen,
                                         ftp_address = ftp_address,
                                         webdav_address = webdav_address,
+                                        icp_address = icp_address,
                                         zserver_threads = zserver_threads,
                                         zodb_cache_size = zodb_cache_size,
                                         zodb_cache_size_bytes = zodb_cache_size_bytes,
@@ -431,7 +465,7 @@ class Recipe:
             fd.close()
 
     def patch_binaries(self, ws_locations):
-        if self.zope2_egg:
+        if not self.zope2_location:
             return
 
         location = self.options['location']
@@ -456,37 +490,39 @@ class Recipe:
         # Patch Windows scripts
         for script_name in ('runzope.bat', ):
             script_path = os.path.join(location, 'bin', script_name)
-            script = open(script_path).read()
-            # This could need some regex-fu
-            lines = [l for l in script.splitlines()
-                     if not l.startswith('@set PYTHON=')]
-            lines.insert(2, '@set PYTHON=%s' % self.options['executable'])
-            script = '\n'.join(lines)
+            if os.path.exists(script_path):
+                script = open(script_path).read()
+                # This could need some regex-fu
+                lines = [l for l in script.splitlines()
+                         if not l.startswith('@set PYTHON=')]
+                lines.insert(2, '@set PYTHON=%s' % self.options['executable'])
+                script = '\n'.join(lines)
 
-            # Use servicewrapper.py instead of calling run.py directly
-            # so that sys.path gets properly set. We used to append
-            # all the eggs to PYTHONPATH in runzope.bat, but after
-            # everything was turned into eggs we exceeded the
-            # environment maximum size for cmd.exe.
-            script = script.replace(
-                "ZOPE_RUN=%SOFTWARE_HOME%\\Zope2\\Startup\\run.py",
-                "ZOPE_RUN=%INSTANCE_HOME%\\bin\\servicewrapper.py"
-                )
-            f = open(script_path, 'w')
-            f.write(script)
-            f.close()
+                # Use servicewrapper.py instead of calling run.py directly
+                # so that sys.path gets properly set. We used to append
+                # all the eggs to PYTHONPATH in runzope.bat, but after
+                # everything was turned into eggs we exceeded the
+                # environment maximum size for cmd.exe.
+                script = script.replace(
+                    "ZOPE_RUN=%SOFTWARE_HOME%\\Zope2\\Startup\\run.py",
+                    "ZOPE_RUN=%INSTANCE_HOME%\\bin\\servicewrapper.py"
+                    )
+                f = open(script_path, 'w')
+                f.write(script)
+                f.close()
         # Patch Windows service scripts
         path =";".join(ws_locations)
         script_name = 'zopeservice.py'
         script_path = os.path.join(location, 'bin', script_name)
-        script = open(script_path).read()
-        script = script.replace(
-                "ZOPE_RUN = r'%s\\Zope2\\Startup\\run.py' % SOFTWARE_HOME",
-                "ZOPE_RUN = r'%s\\bin\\servicewrapper.py' % INSTANCE_HOME"
-                )
-        f = open(script_path, 'w')
-        f.write(script)
-        f.close()
+        if os.path.exists(script_path):
+            script = open(script_path).read()
+            script = script.replace(
+                    "ZOPE_RUN = r'%s\\Zope2\\Startup\\run.py' % SOFTWARE_HOME",
+                    "ZOPE_RUN = r'%s\\bin\\servicewrapper.py' % INSTANCE_HOME"
+                    )
+            f = open(script_path, 'w')
+            f.write(script)
+            f.close()
         script_name = 'servicewrapper.py'
         script_path = os.path.join(location, 'bin', script_name)
         script = """import sys
@@ -504,18 +540,19 @@ if __name__ == '__main__':
         # Add a test.bat that works on Windows
         new_script_path = os.path.join(location, 'bin', 'test.bat')
         script_path = os.path.join(location, 'bin', 'runzope.bat')
-        script = open(script_path).read()
-        # Adjust script to use the right command
-        script = script.replace("@set ZOPE_RUN=%SOFTWARE_HOME%\\Zope2\\Startup\\run.py",
-                                """@set ZOPE_RUN=%ZOPE_HOME%\\test.py
+        if os.path.exists(script_path):
+            script = open(script_path).read()
+            # Adjust script to use the right command
+            script = script.replace("@set ZOPE_RUN=%SOFTWARE_HOME%\\Zope2\\Startup\\run.py",
+                                    """@set ZOPE_RUN=%ZOPE_HOME%\\test.py
 @set ERRLEV=0""")
-        script = script.replace("\"%ZOPE_RUN%\" -C \"%CONFIG_FILE%\" %1 %2 %3 %4 %5 %6 %7",
-                                """\"%ZOPE_RUN%\" --config-file \"%CONFIG_FILE%\" %1 %2 %3 %4 %5 %6 %7 %8 %9
+            script = script.replace("\"%ZOPE_RUN%\" -C \"%CONFIG_FILE%\" %1 %2 %3 %4 %5 %6 %7",
+                                    """\"%ZOPE_RUN%\" --config-file \"%CONFIG_FILE%\" %1 %2 %3 %4 %5 %6 %7 %8 %9
 @IF %ERRORLEVEL% NEQ 0 SET ERRLEV=1
 @ECHO \"%ERRLEV%\">%INSTANCE_HOME%\\testsexitcode.err""")
-        f = open(new_script_path, 'w')
-        f.write(script)
-        f.close()
+            f = open(new_script_path, 'w')
+            f.write(script)
+            f.close()
 
     def install_scripts(self):
         options = self.options
@@ -527,7 +564,7 @@ if __name__ == '__main__':
 
         # Only append the instance home and Zope lib/python in a non-egg
         # environment
-        lib_python = os.path.join(options['zope2-location'], 'lib', 'python')
+        lib_python = os.path.join(self.zope2_location, 'lib', 'python')
         if os.path.exists(lib_python):
             extra_paths.append(os.path.join(location))
             extra_paths.append(lib_python)
@@ -536,40 +573,30 @@ if __name__ == '__main__':
 
         requirements, ws = self.egg.working_set(['plone.recipe.zope2instance'])
 
-        if self.zope2_egg:
-            zc.buildout.easy_install.scripts(
-                [(self.options.get('control-script', self.name),
-                    'plone.recipe.zope2instance.simple', 'main')],
-                ws, options['executable'], options['bin-directory'],
-                extra_paths = extra_paths,
-                arguments = ('\n        [%r]'
-                             % zope_conf
-                             ),
-                )
-        else:
-            zc.buildout.easy_install.scripts(
-                [(self.options.get('control-script', self.name),
-                    'plone.recipe.zope2instance.ctl', 'main')],
-                ws, options['executable'], options['bin-directory'],
-                extra_paths = extra_paths,
-                arguments = ('\n        ["-C", %r]'
-                             '\n        + sys.argv[1:]'
-                             % zope_conf
-                             ),
-                )
-            # The backup script, pointing to repozo.py
-            repozo = options.get('repozo', None)
-            if repozo is None:
-                repozo = os.path.join(options['zope2-location'], 'utilities', 'ZODBTools', 'repozo.py')
+        zc.buildout.easy_install.scripts(
+            [(self.options.get('control-script', self.name),
+                'plone.recipe.zope2instance.ctl', 'main')],
+            ws, options['executable'], options['bin-directory'],
+            extra_paths = extra_paths,
+            arguments = ('\n        ["-C", %r]'
+                         '\n        + sys.argv[1:]'
+                         % zope_conf
+                         ),
+            )
 
-            directory, filename = os.path.split(repozo)
-            if repozo and os.path.exists(repozo):
-                zc.buildout.easy_install.scripts(
-                    [('repozo', os.path.splitext(filename)[0], 'main')],
-                    {}, options['executable'], options['bin-directory'],
-                    extra_paths = [os.path.join(options['zope2-location'], 'lib', 'python'),
-                                   directory],
-                    )
+        # The backup script, pointing to repozo.py
+        repozo = options.get('repozo', None)
+        if repozo is None:
+            repozo = os.path.join(self.zope2_location, 'utilities', 'ZODBTools', 'repozo.py')
+
+        directory, filename = os.path.split(repozo)
+        if repozo and os.path.exists(repozo):
+            zc.buildout.easy_install.scripts(
+                [('repozo', os.path.splitext(filename)[0], 'main')],
+                {}, options['executable'], options['bin-directory'],
+                extra_paths = [os.path.join(self.zope2_location, 'lib', 'python'),
+                               directory],
+                )
 
     def build_package_includes(self):
         """Create ZCML slugs in etc/package-includes
@@ -578,33 +605,41 @@ if __name__ == '__main__':
         sitezcml_path = os.path.join(location, 'etc', 'site.zcml')
         zcml = self.options.get('zcml')
         site_zcml = self.options.get('site-zcml')
+        additional_zcml = self.options.get("zcml-additional")
 
         if site_zcml:
             open(sitezcml_path, 'w').write(site_zcml)
             return
 
         if zcml:
-            if not os.path.exists(sitezcml_path):
-                # Zope 2.9 does not have a site.zcml so we copy the
-                # one out from Five.
-                zope2_location = self.options['zope2-location']
-                skel_path = os.path.join(zope2_location, 'lib', 'python',
-                                         'Products', 'Five', 'skel',
-                                         'site.zcml')
-                shutil.copyfile(skel_path, sitezcml_path)
+            zcml=zcml.split()
 
+        if additional_zcml or zcml:
             includes_path = os.path.join(location, 'etc', 'package-includes')
+
             if not os.path.exists(includes_path):
                 # Zope 2.9 does not have a package-includes so we
                 # create one.
                 os.mkdir(includes_path)
-
-            zcml = zcml.split()
-            if '*' in zcml:
-                zcml.remove('*')
             else:
-                shutil.rmtree(includes_path)
-                os.mkdir(includes_path)
+                if '*' in zcml:
+                    zcml.remove('*')
+                else:
+                    shutil.rmtree(includes_path)
+                    os.mkdir(includes_path)
+
+        if additional_zcml:
+            path=os.path.join(includes_path, "999-additional-overrides.zcml")
+            open(path, "w").write(additional_zcml.strip())
+
+        if zcml:
+            if not os.path.exists(sitezcml_path):
+                # Zope 2.9 does not have a site.zcml so we copy the
+                # one out from Five.
+                skel_path = os.path.join(self.zope2_location, 'lib', 'python',
+                                         'Products', 'Five', 'skel',
+                                         'site.zcml')
+                shutil.copyfile(skel_path, sitezcml_path)
 
             n = 0
             package_match = re.compile('\w+([.]\w+)*$').match
@@ -638,6 +673,7 @@ if __name__ == '__main__':
                     % (package, filename)
                     )
 
+
 # Storage snippets for zope.conf template
 file_storage_template="""
     # FileStorage database
@@ -656,8 +692,9 @@ demo_storage_template="""
 rel_storage_template="""
     %%import relstorage
     <relstorage>
+%(rs_opts)s
         <%(type)s>
-%(opts)s
+%(db_opts)s
         </%(type)s>
     </relstorage>
 """
@@ -748,6 +785,13 @@ ftp_server_template = """
 </ftp-server>
 """
 
+icp_server_template = """
+<icp-server>
+  # valid key is "address"
+  address %s
+</icp-server>
+"""
+
 webdav_server_template = """
 <webdav-source-server>
   # valid keys are "address" and "force-connection-close"
@@ -803,6 +847,7 @@ verbose-security %(verbose_security)s
 
 %(ftp_address)s
 %(webdav_address)s
+%(icp_address)s
 
 <zodb_db main>
     # Main database
